/*****************************************************
*
*	GPS Watch code created by:
* 
*	Joel Heck and Kyle Kulas
*
*	November 2013
*
*	ECE4760 Fall 2013
*
*****************************************************/

#define debug 1 //Set to 0 to disable debugging prints
#include "GPS_watch.h"

int waypt_index = 1;	// Index of list of waypoints
uint16_t dist_rem = 0;	// Distance remaining to next waypoint (m)
int16_t bear = 0;		// Bearing to next waypoint (0-360 deg)
int8_t prev_fix = -1;	// State of previous GPS fix
uint8_t run_done = 0;	// Indicates whether or not run has been completed
uint8_t run_start = 0;	// Indicates whether or not run has started
char lcd_buffer[17];	// LCD buffer

/* UART string */
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

int main(void){
	menu_selection=1;
	/* Initialize interrupts, ports, peripherals */
	initialize();
	
	/* Run scheduler */
	init_scheduler();
	while(1){
		scheduler();
	}
}


/*****************************************************
* Tasks and subroutines								 *
*****************************************************/

/* Initialize interrupts, ports, peripherals */
void initialize(void){
	int i;

	/* UART Terminal initialization */
	uart_init();
	//stdout = stdin = stderr = &uart_str;
	fprintf(&uart_str,"Hello!\r\n");	

	/* Initialize LCD */
	LCDinit();
	LCDcursorOFF(); 						//disable cursor
    LCDclr();   							//clear the display  

	/* Set up timer 0 for 1 mSec timebase */
	OCR0A = 249;			//set the compare re to 250 time ticks 
	TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR   
	TCCR0B= 3;				//set prescalar to divide by 64 
	TCCR0A= (1<<WGM01);		//turn on clear-on-match 

	/* Setup pin directions */
	/* A4,A5 Vibration motors (output) */
	DDRA = 0b00110000;
	/* A0-A3 input buttons (pull up) */
	PORTA = 0b00001111;
	
	// Initialize EEPROM
	if (eeprom_read_byte((uint8_t*)EEPROM_SAVE) != 'S') {
		eeprom_write_byte((uint8_t*)EEPROM_SAVE, 'S');
		eeprom_write_float((float*)EEPROM_TIME,0.0);
		eeprom_write_float((float*)EEPROM_DIST,0.0);	
	}

	/* Initialize GPS structure */
	init_gps(&raw);
	num_routes=0;

	/* Enable interrupts globally */
	sei();

	LCD_print("Initializing SD",0);
	
	/* Initialize SD card */
	/* Add back in if hardware is reliable 
	if(!(PINB & 0x08)){ // Chip Detect on B.3
		LCD_print("No SD Card",0);
	}

	while(!(PINB & 0x08)){
		_delay_ms(1000);
		fprintf(&uart_str,"PIN: %i\r\n",(PINB & 8));
	}
	_delay_ms(1000); */
	init_fs();
	
	/* Get list of routes from SD card */
	scancard(route_list,&num_routes,MAX_ROUTES);
	
	/* Remove file extension */
	for(i=0;i<num_routes;i++){
		route_list[i][strlen(route_list[i])-1]=0;
		route_list[i][strlen(route_list[i])-1]=0;
		route_list[i][strlen(route_list[i])-1]=0;
		route_list[i][strlen(route_list[i])-1]=0;
	}
	#ifdef debug
	fprintf(&uart_str,"routes %i\r\n", num_routes);
	
	for(i=0;i<num_routes;i++){
		fprintf(&uart_str,"%s\r\n",route_list[i]);
	}
	#endif

	/* Initial LCD menu text */
	LCD_print("Menu:",0);
	LCD_print("   Navigation ->",1);
}

/************************************************************/

/* Prints str to the selected line on the LCD left justified
	and padded to 16 characters */
void LCD_print(char* str, uint8_t line){
	char buffer[17];
	LCDGotoXY(0,line);
	sprintf(buffer,"%-16.16s",str);
	LCDstring((uint8_t*) buffer,16);
}

/************************************************************/

/* Main Menu task */
/* Returns 0 while running, 1 to switch to navigation,
	2 to switch to statistics, 3 to switch to settings */
void LCD_menu_update(uint8_t); //Below
uint8_t menu(uint8_t init){
	uint8_t status=0;
	/* Initialize main menu */
	if(init){
		menu_selection=1;
		LCD_print("Menu:",0);
		LCD_menu_update(1);
		return 0;
	}
	/* Wait on user input */
	debounce();
	if(button_flag){
		switch (butnum){
			case LEFT:
				if(menu_selection>1){ // Scroll left if possible
					menu_selection--;
					LCD_menu_update(menu_selection);
				}
				break;
			case RIGHT:
				if(menu_selection<3){ // Scroll right if possible
					menu_selection++;
					LCD_menu_update(menu_selection);
				}
				break;
			case SELECT:
				status = menu_selection; // Return next program state to scheduler
				break;
			default: //Don't care about Back
				break;
		}
		button_flag = FALSE;
	}
	return status;
}

/* Updates LCD based on left and right buttons in main menu */
void LCD_menu_update(uint8_t sel){
	switch(sel){
		case 1://Left end
		LCD_print("   Navigation ->",1);
		break;
		case 2:
		LCD_print("<-   Logger   ->",1);
		break;
		case 3://Right end
		LCD_print("<- Statistics",1);
		break;
	}
}

/************************************************************/

/* Logging menu
 * Opens new route file
 * Brings up watch and current location on display
 * Waits on user input. Select adds current location to route
 */
uint8_t logger(uint8_t init){
	char line[34];
	if(init){
		if (num_routes==MAX_ROUTES){ //If route list full display error
			LCD_print("Route list full",0);
			LCD_print("",1);
			_delay_ms(1000);
			return 1;
		}
		// Open file
		openFile("user","txt",log_name);
		// Write header
		writeLine("type,latitude,longitude,name,desc\r\n");
		LCD_print("Press Sel to log",0); // Update LCD
		LCD_print("a new point",1);
		_delay_ms(1000);
		return 0;
	}
	debounce(); //Wait on user input
	if(button_flag){
		switch (butnum){
			case SELECT:
				LCD_print("Waypoint saved", 0);
				_delay_ms(500);
				// Write current location to string then write to file
				sprintf(line,"T,%f,%f,,\r\n",raw.latitude,raw.longitude);
				writeLine(line);
				break;
			case BACK://close file, update list of routes
				closeFile();
				sprintf(route_list[num_routes],log_name);
				num_routes++;
				return 1; // return to main menu
			break;
		}
		button_flag= FALSE;
	}
	return 0;
}

/*
 * Save the current position to a text file on the SD card
 */
void log_waypt(void) {

	// Parse the GPS data and store the state of the fix in status
	int status = get_gps();
	
	switch (status) {
		// Good GPS fix and data available
		case 0:			
			// Determine (N/S) and (E/W) printout
			if ((raw.latitude > 0) && (raw.longitude > 0))
				sprintf(lcd_buffer, "%2.3fN %3.3fE", raw.latitude, 
					raw.longitude);
			else if ((raw.latitude > 0) && (raw.longitude < 0))
				sprintf(lcd_buffer, "%2.3fN %3.3fW", raw.latitude, 
					raw.longitude);
			else if ((raw.latitude < 0) && (raw.longitude > 0))
				sprintf(lcd_buffer, "%2.3fS %3.3fE", raw.latitude, 
					raw.longitude);
			else
				sprintf(lcd_buffer, "%2.3fS %3.3fW", raw.latitude, 
					raw.longitude);
						
			// Print distance and time on LCD
			LCD_print(lcd_buffer, 0);
			print_time(2);
			break;
		case -2:
			if (prev_fix != 0) {
				LCD_print("Waiting for GPS", 0);
				LCD_print("fix...", 1);
			} else {
				print_time(2);	
			}
			break;
		case -1:
			// Do not change LCD message if just had fix
			if (prev_fix != 0) {
				LCD_print("Waiting for GPS", 0);
				LCD_print("fix...", 1);
			}
			break;
	} // end switch

	prev_fix = status;	// Previous state of fix
}

/************************************************************/

/* Statistics menu */
void LCD_stats_update(uint8_t);
uint8_t stats(uint8_t init){
	uint8_t status=0;
	if(init) { // Read total distance and time from EEPROM into memory
		total_dist = eeprom_read_float((float*)EEPROM_DIST);
		total_time = eeprom_read_float((float*)EEPROM_TIME);
		menu_selection=1;
		LCD_stats_update(1); // Display first statistic
		return 0;
	}
	debounce(); // Wait on user input
	if(button_flag){
		switch (butnum){
			case LEFT:
			if(menu_selection>1){
				menu_selection--;
				LCD_stats_update(menu_selection);
			}
			break;
			case RIGHT:
			if(menu_selection<3){
				menu_selection++;
				LCD_stats_update(menu_selection);
			}
			break;
			case SELECT:
			if(menu_selection==3){
				eeprom_write_float((float*)EEPROM_DIST,0);
				eeprom_write_float((float*)EEPROM_TIME,0);
				total_dist=0;
				total_time=0;
				LCD_print("  Statistics",0);
				LCD_print("   Cleared",1);
				status = 1;
			}
			break;
			case BACK:
			status = 1;
			break;
		}
		button_flag = FALSE;
	}
	return status;
}

void LCD_stats_update(uint8_t sel){
	char buff[16];
	double min;
	double sec;
	switch(sel){
		case 1:
			LCD_print("Total Distance:",0);
			sprintf(buff,"%.2f km",total_dist);
			LCD_print(buff,1);
		break;
		case 2:
			LCD_print("Total Time:",0);
			min = (total_time-floor(total_time))*60;
			sec = (min-floor(min))*60;
			//fprintf(&uart_str,"Time: %f\r\n",sec);
			sprintf(buff,"%.3i:%.2i:%.2i",(int) total_time,(int) min,(int) sec);
			LCD_print(buff,1);
		break;
		case 3:
		LCD_print("Reset Stats",0);
		LCD_print("Press Select",1);
		break;
	}
}

/************************************************************/

/* Route selection menu 
 * Called when navigation is selected in main menu
 * Lists routes stored on device
 */
uint8_t route_select(uint8_t init){
	uint8_t status = 0;
	int i = 0;
	if(init){
		if(num_routes==0){
			LCD_print("No routes found",0);
		} else {
			LCD_print("Select a Route",0);
			LCD_print(route_list[0],1);
		}
		menu_selection=0;
		return 0;
	}
	debounce();
	if(button_flag){
		switch (butnum){
			case LEFT: //Left
				if(menu_selection>0){
					menu_selection--;
					LCD_print(route_list[menu_selection],1);
				}
			break;
			case RIGHT: //Right
				if(menu_selection<num_routes-1){
					menu_selection++;
					LCD_print(route_list[menu_selection],1);
				}
			break;
			case SELECT: //Select
				status = 2;
				LCD_print("Loading ", 0);
				LCD_print("waypoints...", 1);
				load_waypoints(route_list[menu_selection],waypt_lat,waypt_long);
				#ifdef debug
				while(waypt_lat[i]!=-1 && i<100){
					fprintf(&uart_str,"Point %i\r\n",i+1);
					fprintf(&uart_str,"Lat: %f\r\n",waypt_lat[i]);
					fprintf(&uart_str,"Long: %f\r\n",waypt_long[i]);
					i++;
				}
				#endif
			break;
			case BACK: //Back
				status = 1;
			break;
		}
		button_flag = FALSE;
	}
	return status;
}

/* Runs in navigation mode and exits navigation mode on back button press */
uint8_t nav_button(void){
	uint8_t ret=0;
	debounce();
	if(button_flag){
		if(butnum==BACK){ // Clear navigation parameters
			PORTA &= ~(0x01<<4);
			PORTA &= ~(0x01<<5);	// Turn motors off
			raw.elapsed_time=0;
			raw.dist_rem=0;
			raw.dist_done=0;
			waypt_index = 1;
			run_start = 0;
			run_done = 0;
			ret=1;
		}
		button_flag = FALSE;
	}
	return ret;
}

/*
 * Performs calculations to navigate the user between waypoints. When a user
 * is close enough to a waypoint, the waypoint is updated to the next waypoint
 * in the list. It provides guidance to the user by vibrating a motor in the 
 * direction of the next waypoint and displaying information on an LCD display
 */
void navigation(void){
	
	// Parse the GPS data and store the state of the fix in status
	int gps_status = get_gps();
	int8_t turn = 0;
	
	if (waypt_lat[waypt_index] != -200 && run_start) {
		raw.elapsed_time++;	// Increment elapsed time of run
	}
	
	switch (gps_status) {
		// Good GPS fix and data available
		case 0: 
			// Quit if the route is complete, otherwise navigate
			if (waypt_lat[waypt_index] == -200) {
				// Only save accumulated dist and time once
				if (run_done == 0) {
					PORTA |= (0x01<<4) | (0x01<<5);	// Turn motors on
					run_done++;
					raw.dist_done = raw.dist_done/1000.0;	// Convert to km
					
					// Read saved accumulated distance (km)
					total_dist = eeprom_read_float(
						(float*)EEPROM_DIST);
					total_dist += raw.dist_done;
					// Save new accumulated distance
					eeprom_write_float((float*)EEPROM_DIST,total_dist); 
					
					// Read saved accumulated time (hrs)
					total_time = eeprom_read_float(
						(float*)EEPROM_TIME);
					total_time += ((double)raw.elapsed_time/3600.0);
					// Save new accumulated time
					eeprom_write_float((float*)EEPROM_TIME,total_time);
					_delay_ms(200);
				}

				PORTA &= ~(0x01<<4);
				PORTA &= ~(0x01<<5);
				LCD_print("Run Complete!", 0);
				print_time(1);
			} else if (run_start){
				// Find distance and bearing between user and next waypoint
				dist_rem = approx_dist_to_waypt(raw.latitude,raw.longitude,
					waypt_lat[waypt_index],waypt_long[waypt_index]);
				
				// Check that new remaining distance is reasonable	
				uint16_t temp_dist = approx_dist_to_waypt(raw.prev_lat, 
					raw.prev_lon, raw.latitude, raw.longitude);
				if (temp_dist < 100) {
					raw.dist_done += temp_dist;
					raw.dist_rem = dist_rem;
				} else {
					dist_rem = raw.dist_rem;
				}
				
				bear = bearing_to_waypt(raw.latitude,raw.longitude,
					waypt_lat[waypt_index],waypt_long[waypt_index]);
				
				// Print distance and time on LCD
				sprintf(lcd_buffer, "Dist-%u Spd-%d", dist_rem, raw.speed);
				LCD_print(lcd_buffer, 1);
				print_time(0);
				
				// Advance to next waypoint if within a given distance of the 
				// current waypoint 
				if (dist_rem < (uint16_t)NEAR_DIST) {
					waypt_index++;
					
					// Don't calculate bearing if run complete
					if (waypt_lat[waypt_index] != -200) {
						bear = bearing_to_waypt(raw.latitude,raw.longitude,
							waypt_lat[waypt_index],waypt_long[waypt_index]);
						turn = dir_to_waypt(raw.heading, bear, 1);
					} else {
						PORTA &= ~(0x01<<4);
						PORTA &= ~(0x01<<5); 	// Turn off motors
					}
				} else if (dist_rem < (uint16_t)INIT_DIST) {	// Give  
											//upcoming turn indication if close
					
					// Don't calculate bearing if run complete
					if (waypt_lat[waypt_index + 1] != -200) {
						bear = bearing_to_waypt(raw.latitude,raw.longitude,
							waypt_lat[waypt_index + 1],
							waypt_long[waypt_index + 1]);
						turn = dir_to_waypt(raw.heading, bear, 1);
					} else {
						PORTA &= ~(0x01<<4);
						PORTA &= ~(0x01<<5); 	// Turn off motors
					}
				} else {
					turn = dir_to_waypt(raw.heading, bear, 0);
				}
				
				// Don't engage motors if too slow or just started (inaccurate
				// bearing measurements
				if (raw.elapsed_time > 10 && raw.speed > 5) {	
					switch (turn) {
						case -1:
							PORTA |= (0x01<<4);		// Turn on left motor
							PORTA &= ~(0x01<<5); 	// Turn off right motor
							break;
						case 0:
							PORTA |= (0x01<<4) | (0x01<<5);	// Turn on both 
															// motors
							break;
						case 1:
							PORTA &= ~(0x01<<4); 	// Turn off left motor
							PORTA |= (0x01<<5);		// Turn on right motor
							break;
						case 2:
							PORTA &= ~(0x01<<4);
							PORTA &= ~(0x01<<5); 	// Turn off motors
							break;
					} // end switch
				}
			} else if(!run_start) {
				run_start = 1;	
			}
			raw.prev_lat = raw.latitude;
			raw.prev_lon = raw.longitude;
			break;
		case -2: 
			// Display wait message while looking for sats
			if (!run_start) {
				LCD_print("Waiting for GPS", 0);
				LCD_print("fix...", 1);
			} else {
				print_time(0);
				LCD_print("Bad fix...", 1);
			}
			break;
		case -1:
			// Do not change LCD message if just had fix
			if (prev_fix != 0) {
				LCD_print("Waiting for GPS", 0);
				LCD_print("fix...", 1);
			}
			break;
	} // end switch
	prev_fix = gps_status;	// Previous state of fix

} // end navigation()

/*
 * Calculates the equirectangular approximation of the distance between two 
 * (latitude, longitude) positions in meters. This approximation should work 
 * for distances of at least 100 km, which is far greater than necessary for
 * this project.
 *
 * lat: latitude in decimal degrees of first position
 * lon: longitude in decimal degrees of first position
 * next_lat: latitude in decimal degrees of second position
 * next_lon: longitude in decimal degrees of second position
 *
 * return: integer distance from the first position to the second position
 * 		   in meters
 */
uint16_t approx_dist_to_waypt(double lat, double lon, double next_lat, double next_lon) {
	double x = (next_lon-lon)*DEGRAD*cos((lat+next_lat)*DEGRAD/2);
	double y = (next_lat-lat)*DEGRAD;
	
	return (uint16_t)(sqrt(x*x + y*y)*ERADIUS);
}

/*
 * Finds the bearing between two (latitude, longitude) positions in degrees
 *
 * lat: latitude in decimal degrees of first position
 * lon: longitude in decimal degrees of first position
 * next_lat: latitude in decimal degrees of second position
 * next_lon: longitude in decimal degrees of second position
 *
 * return: integer bearing from the first position to the second position
 * 		   in degrees
 */
int16_t bearing_to_waypt(double lat, double lon, double next_lat, double next_lon) {
	double dLong = next_lon-lon;
	double dy = sin(dLong*DEGRAD)*cos(next_lat*DEGRAD);
	double dx = cos(lat*DEGRAD)*sin(next_lat*DEGRAD)-
		sin(lat*DEGRAD)*cos(next_lat*DEGRAD)*cos(dLong*DEGRAD);
	int16_t result = (int16_t)(atan2(dy,dx)*57.3);	// 57.3 = 180/pi
	
	if (result > 0) 
		return result;
	else
		return 360+result;
}

/*
 * Finds the direction to turn to reach the active waypoint if off-track or 
 * transitioning to the next waypoint. Only gives an indication to turn if 
 * the user's bearing is more than 60 degrees from the expected bearing
 *
 * actual: the actual bearing of the user in integer degrees (0-360)
 * expected: the bearing from the user's current position to the active 
 * 		waypoint in integer degrees (0-360)
 * next: an integer (0 or 1) to indicate if transitioning to a new waypoint
 *
 * return: an integer giving the direction to turn (if any) or an error
 * 		-2 = input out of bounds, -1 = turn left, 0 = straight, 1 = turn right
 * 		2 = do nothing (on track)
 */
int8_t dir_to_waypt(int16_t actual, int16_t expected, uint8_t next) {
	if (actual > 360 || expected > 360 || actual < 0 || expected < 0)
		return -2;	// Error (input must be in range [0,360])

	int16_t diff = expected - actual;
	int16_t absD = (diff < 0) ? -diff : diff;

	if (absD > TURN_IND || next == 1) {
		if (absD == 0)
			return 0;
		else if (absD < 180)
			return (diff < 0) ? -1 : 1;
		else if (absD == 180)
			return -1;	// Arbitrarily chose turn left
		else if (expected > actual) {
			if (((actual+360)-expected) > TURN_IND)	
				return -1;
		}
		else {
			if (((expected+360)-actual) > TURN_IND)	
				return 1;
		}
	}

	return 2;
}

/*
 * Formats and prints the elapsed time of run or the current local time to the
 * LCD display. Depending on the argument, prints other information, such as 
 * altitude, waypoint number, or distance completed in the run.
 *
 * line: an integer to determine which information to print
 */
void print_time(uint8_t line) {
	uint8_t hr;
	uint8_t min;
	uint8_t sec;
	uint32_t time = raw.elapsed_time;

	if (line != 2) {	// Convert elapsed time (sec) to hh:mm:ss
		hr = (uint8_t)(time / 3600UL);
		time -= ((uint32_t)hr * 3600UL);
		min = (uint8_t)(time / 60UL);
		sec = (uint8_t)(time - (uint32_t)min * 60UL);
	} else {	// Pull current time (hour and minute) 
		hr = raw.hour;
		min = raw.minute;
		sec = 0;
	}

	// Char arrays to hold formatted numbers
	char strhr[3];
	char strmin[3];
	char strsec[3];

	// Change hour to two digit number
	if (hr < 10)
		sprintf(strhr, "0%d", hr);
	else
		sprintf(strhr, "%d", hr);
	// Change minute to two digit number
	if (min < 10)
		sprintf(strmin, "0%d", min);
	else
		sprintf(strmin, "%d", min);
	// Change second to two digit number
	if (sec < 10)
		sprintf(strsec, "0%d", sec);
	else
		sprintf(strsec, "%d", sec);

	if (line == 0) 	// Print waypoint number and elapsed time (during run)
		sprintf(lcd_buffer, "W-%d T-%s:%s:%s", waypt_index, strhr, strmin, 
			strsec); 
	else if (line == 1)		// Print completed distance and time (after run)
		sprintf(lcd_buffer, "%.1fkm %s:%s:%s", raw.dist_done, strhr, strmin, 
			strsec);
	else if (line == 2) {	// Print altitude and local time (logging)
		sprintf(lcd_buffer, "Alt-%d %s:%s", raw.altitude, strhr, strmin);
		line = 1;
	}

	LCD_print(lcd_buffer, line);
}

