#include "sd.h"
#include "shared.h"
//#define DEBUG

/* Initializes filesystem on the SD card */
uint8_t init_fs() {
	DSTATUS dstatus;
	uint8_t err = 0;
    
	dstatus = disk_initialize(0);

    #ifdef DEBUG
	fprintf(&uart_str, "Disk Initiliazation: %d\n\r", dstatus);
	if(dstatus == STA_NOINIT) fprintf(&uart_str, "STA_NOINIT ");
	else if(dstatus == STA_NODISK) fprintf(&uart_str, "STA_NODISK ");
	else if(dstatus == STA_PROTECT) fprintf(&uart_str, "STA_PROTECT ");
    #endif
	
    if(dstatus == 0) diskInitFlag = 1; //Disk has been Initialized
	
	FRESULT result;
	//init pointer
	fsPtr = &fs;
	filePtr = &file;
	//set up file system
	result = f_mount(0, fsPtr);
    #ifdef DEBUG
	fprintf(&uart_str, "Initiating filesystem mount: %d\n\r", result);
    #endif    

    if(result == 0) diskMountFlag = 1; //Disk has been mounted
	if(!diskInitFlag || !diskMountFlag) err=1;
	return err;
}

//Opens file in R/W
void openFile(char* fName, char* ext, char* name) {
	FRESULT fileResult;
	char filePath[16];
	unsigned int fileCounter = 0;
	
	if(diskInitFlag && diskMountFlag) {
	//Open file adding X for files of same name (X is a number)
	do {
		if(fileCounter>0){
			snprintf(filePath, 16, "%s%i.%s", fName, fileCounter,ext);
			sprintf(name,"%s%i",fName,fileCounter);
		} else {
			snprintf(filePath, 16, "%s.%s", fName,ext);
			sprintf(name,"%s",fName);
		}
		fileResult = f_open(filePtr, filePath,  FA_WRITE | FA_CREATE_NEW); //FA_WRITE | FA_CREATE_NEW
	} while (fileResult == FR_EXIST && ++fileCounter < 1000);
	if (fileResult) {
		#ifdef DEBUG
		fprintf(&uart_str, "File creation FAILED: %d\r\n", fileResult);
		printFRcode(fileResult);
		#endif
	}
	else {
		#ifdef DEBUG
		fprintf(&uart_str, "File Opened: %s \r\n", filePath);
		#endif
		openFlag = 1;
	}
	}
    else {
        #ifdef DEBUG
        fprintf(&uart_str, "Disk not mounted/initialized can't open file\n\r");
        #endif
    }
}

//Closes the file
void closeFile(void) {
    //only try to close if the file has been opened
	FRESULT fileResult;
	if(openFlag) {
        fileResult = f_close(&file);
        if(fileResult) {
            #ifdef DEBUG
            fprintf(&uart_str, "File Closing Failed: ");
            printFRcode(fileResult);
            #endif
        }
        else {
            #ifdef DEBUG
            fprintf(&uart_str, "File Close.\n\r");
            #endif
            openFlag = 0;
        }
    }
}


void writeLine(char* line) {
    FRESULT fileResult;
    unsigned int length;
    unsigned int bytesWrote;
	
    //Write string to file
    length = strlen(line);
	fileResult = f_write(filePtr, line, strlen(line), &bytesWrote);
	
	#ifdef DEBUG
	if (fileResult | (bytesWrote != length)) {
		fprintf(&uart_str, "Log write error. Result %d, Tried %d, Wrote %d\r\n", fileResult, length, bytesWrote);
	}
	else {
		fprintf(&uart_str, "Log write succeeded.\r\n");
	}
	#endif
	//sync the filesystem (flush buffers) so it will power down cleanly
	f_sync(filePtr);
}

DWORD get_fattime()
{
	return      ((DWORD)(2012 - 1980) << 25)
            | ((DWORD)5 << 21)
            | ((DWORD)(day+1) << 16)
            | ((DWORD)hours << 11)
            | ((DWORD)mins << 5)
            | ((DWORD)(millis >> 10) >> 1);
    // div 1024 is close enough to div 1000 since the time is
    // pretty made up in this case anyway   
}

//Print out file result code
void printFRcode(FRESULT fResult) {
	switch(fResult) {
		case FR_OK:
			fprintf(&uart_str, "FR OK\n\r");
			break;
		case FR_DISK_ERR:
			fprintf(&uart_str, "FR Disk Error\n\r");
			break;
		case FR_INT_ERR:
			fprintf(&uart_str, "FR_INT_ERR\n\r");
			break;
		case FR_NOT_READY:
			fprintf(&uart_str, "FR_NOT_READY\n\r");
			break;
		case FR_NO_FILE:
			fprintf(&uart_str, "FR_NO_FILE\n\r");
			break;			
		case FR_NO_PATH:
			fprintf(&uart_str, "FR_NO_PATH\n\r");
			break;			
		case FR_INVALID_NAME:
			fprintf(&uart_str, "FR_INVALID_NAME\n\r");
			break;	
		case FR_DENIED:
			fprintf(&uart_str, "FR_DENIED\n\r");
			break;			
		case FR_EXIST:
			fprintf(&uart_str, "FR_EXIST\n\r");
			break;			
		case FR_INVALID_OBJECT:
			fprintf(&uart_str, "FR_INVALID_OBJECT\n\r");
			break;	
		case FR_WRITE_PROTECTED:
			fprintf(&uart_str, "FR_WRITE_PROTECTED\n\r");
			break;	
		case FR_INVALID_DRIVE:
			fprintf(&uart_str, "FR_INVALID_DRIVE\n\r");
			break;	
		case FR_NOT_ENABLED:
			fprintf(&uart_str, "FR_NOT_ENABLED\n\r");
			break;		
		case FR_NO_FILESYSTEM:	
			fprintf(&uart_str, "FR_NO_FILESYSTEM\n\r");
			break;
		case FR_MKFS_ABORTED:	
			fprintf(&uart_str, "FR_MKFS_ABORTED\n\r");
			break;
		case FR_TIMEOUT:
			fprintf(&uart_str, "FR_TIMEOUT\n\r");
			break;			
		default:
			fprintf(&uart_str, "Unknown FR code\n\r");
			break;
	}
}

/* Returns list of files in the root in an arrayof file
	names. Filenames are 14 char or less and the number
	of rows in file_array must equal array_size 
	Public shared function */
void scancard(
	char file_array[][14],
	uint8_t* routes,
	uint8_t array_size)
{
	file_list("",file_array, routes, array_size);
}

/* Returns list of files in the specified path in an array
	of file names. Filenames are 14 char or less and the 
	number of rows in file_array must equal array_size 
	Private function */
FRESULT file_list (
    char* path,       /* Path to scan */
	char file_array[][14], /* Pointer to list of files */
	uint8_t* num_files,	/* Number of files listed */
	uint8_t array_size /* Maximum number of files to scan */)
{
    FRESULT res;
    FILINFO fno;
    DIR dir;
    int i;
	int j;
    char *fn;   /* This function is assuming non-Unicode cfg. */

    res = f_opendir(&dir, path);                       /* Open the directory */
    if (res == FR_OK) {
        i = strlen(path);
        for (j=0;j<array_size;j++) {
            res = f_readdir(&dir, &fno);                   /* Read a directory item */
            if (res != FR_OK || fno.fname[0] == 0) break;  /* Break on error or end of dir */
            if (fno.fname[0] == '.') continue;             /* Ignore dot entry */
            fn = fno.fname;
            if (!(fno.fattrib & AM_DIR)) {                    /* It is not a directory */                                      /* It is a file. */
                sprintf(file_array[j],fn);
				num_files[0]++;
            }
        }
    }
    return res;
}


void load_waypoints (char* fname,double* way_lat, double* way_long){
	FIL fil;       /* File object */
    char line[82]; /* Line buffer */
    FRESULT fr;    /* FatFs return code */
	char fname_full[14];
	char *token;
	uint8_t count = 0;

	sprintf(fname_full,"%s.txt",fname);
	fr = f_open(&fil, fname_full, FA_READ);
	/* Skip first line */
	f_gets(line, sizeof line, &fil);
	/* Read following lines and extract waypoints */
    while (f_gets(line, sizeof line, &fil) && count<100){
		token = strtok(line, ","); // First is "T"
		token = strtok(NULL, ","); // Second is Latitude
		way_lat[count]= atof(token);
		token = strtok(NULL, ","); // Third is Longitude
		way_long[count]= atof(token);
		count++;
	}
	way_lat[count]=way_long[count] = -200; 	// -200 to not conflict with 
										   	// possible data
	/* Close the file */
    f_close(&fil);
}
